import argparse

from tabulate import tabulate

from xklb import consts, db, player, usage, utils
from xklb.history import create
from xklb.utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        "library history",
        usage=usage.history,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--frequency",
        "--freqency",
        "-f",
        metavar="frequency",
        default="monthly",
        const="monthly",
        type=str.lower,
        nargs="?",
        help="One of: %(choices)s (default: %(default)s)",
    )

    parser.add_argument("--sort", "-u", nargs="+", help=argparse.SUPPRESS)
    parser.add_argument("--where", "-w", nargs="+", action="extend", default=[], help=argparse.SUPPRESS)
    parser.add_argument("--include", "-s", "--search", nargs="+", action="extend", default=[], help=argparse.SUPPRESS)
    parser.add_argument("--exclude", "-E", "-e", nargs="+", action="extend", default=[], help=argparse.SUPPRESS)
    parser.add_argument("--duration", "-d", action="append", help=argparse.SUPPRESS)
    parser.add_argument("--limit", "-L", "-l", "-queue", "--queue")

    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("--db", "-db", help=argparse.SUPPRESS)

    parser.add_argument("database")
    parser.add_argument(
        "facet",
        metavar="facet",
        type=str.lower,
        default="watching",
        const="watching",
        nargs="?",
        help="One of: %(choices)s (default: %(default)s)",
    )

    args = parser.parse_args()

    args.facet = utils.partial_startswith(args.facet, consts.time_facets)
    args.frequency = utils.partial_startswith(args.frequency, consts.frequency)

    if args.db:
        args.database = args.db
    args.db = db.connect(args)
    log.info(utils.dict_filter_bool(args.__dict__))

    return args


def print_history(tbl):
    utils.col_duration(tbl, "duration_sum")
    utils.col_duration(tbl, "duration_avg")
    utils.col_naturalsize(tbl, "size_sum")
    utils.col_naturalsize(tbl, "size_avg")
    tbl = utils.list_dict_filter_bool(tbl)
    print(tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False))


def print_recent(tbl, time_column=None):
    utils.col_duration(tbl, "duration")
    utils.col_duration(tbl, "playhead")
    if time_column:
        utils.col_naturaltime(tbl, time_column)
    tbl = [{"title_path": "\n".join(utils.concat(d["title"], d["path"])), **d} for d in tbl]
    tbl = [{k: v for k, v in d.items() if k not in ("title", "path")} for d in tbl]

    tbl = utils.col_resize(tbl, "title_path", 40)
    tbl = utils.col_resize(tbl, "duration", 5)
    tbl = utils.col_resize(tbl, "playhead", 5)
    tbl = utils.list_dict_filter_bool(tbl)
    print(tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False))


def recent_media(args, time_column):
    m_columns = db.columns(args, "media")
    query = f"""
    SELECT
        path
        {', title' if 'title' in m_columns else ''}
        {', duration' if 'duration' in m_columns else ''}
        {', subtitle_count' if 'subtitle_count' in m_columns else ''}
        , {time_column}
    FROM media
    WHERE coalesce({time_column}, 0)>0
    ORDER BY {time_column} desc
    LIMIT {args.limit or 5}
    """
    return list(args.db.query(query))


def history() -> None:
    args = parse_args()
    create(args)

    m_columns = args.db["media"].columns_dict

    if args.facet.startswith("watching"):
        print("Partially watched:")
        tbl = player.historical_usage(args, args.frequency, "time_played", "and coalesce(play_count, 0)=0")
        print_history(tbl)
        query = f"""WITH m as (
                SELECT
                    SUM(CASE WHEN h.done = 1 THEN 1 ELSE 0 END) play_count
                    , MIN(h.time_played) time_first_played
                    , MAX(h.time_played) time_last_played
                    , FIRST_VALUE(h.playhead) OVER (PARTITION BY h.media_id ORDER BY h.time_played DESC) playhead
                    , path
                    {', title' if 'title' in m_columns else ''}
                    {', duration' if 'duration' in m_columns else ''}
                    {', subtitle_count' if 'subtitle_count' in m_columns else ''}
                FROM media m
                LEFT JOIN history h on h.media_id = m.id
                WHERE coalesce(time_deleted, 0) = 0
                GROUP BY m.id, m.path
            )
            SELECT *
            FROM m
            WHERE 1=1
                and coalesce(playhead, 0) > 60
                and coalesce(play_count, 0) = 0
            ORDER BY time_last_played desc, playhead desc
            LIMIT {args.limit or 5}
        """
        tbl = list(args.db.query(query))
        print_recent(tbl, "time_last_played")

    elif args.facet.startswith("watched"):
        print("Finished watching:")
        tbl = player.historical_usage(args, args.frequency, "time_played", "and coalesce(play_count, 0)>0")
        print_history(tbl)
        query = f"""WITH m as (
                SELECT
                    SUM(CASE WHEN h.done = 1 THEN 1 ELSE 0 END) play_count
                    , MIN(h.time_played) time_first_played
                    , MAX(h.time_played) time_last_played
                    , FIRST_VALUE(h.playhead) OVER (PARTITION BY h.media_id ORDER BY h.time_played DESC) playhead
                    , path
                    {', title' if 'title' in m_columns else ''}
                    {', duration' if 'duration' in m_columns else ''}
                    {', subtitle_count' if 'subtitle_count' in m_columns else ''}
                FROM media m
                LEFT JOIN history h on h.media_id = m.id
                GROUP BY m.id, m.path
            )
            SELECT *
            FROM m
            WHERE coalesce(play_count, 0)>0
            ORDER BY time_last_played desc, path
            LIMIT {args.limit or 5}
        """
        tbl = list(args.db.query(query))
        print_recent(tbl, "time_last_played")

    else:
        print(f"{args.facet.title()} media:")
        tbl = player.historical_usage(args, args.frequency, f"time_{args.facet}")
        print_history(tbl)
        tbl = recent_media(args, f"time_{args.facet}")
        print_recent(tbl, f"time_{args.facet}")


if __name__ == "__main__":
    history()



import argparse, os
from copy import deepcopy
from pathlib import Path
from typing import Dict, List

import humanize
from tabulate import tabulate

from xklb import consts, db, usage, utils
from xklb.utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="library disk_usage",
        usage=usage.disk_usage,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--sort-by", "--sort", "-u")
    parser.add_argument("--limit", "-L", "-l", "-queue", "--queue", default="4000")
    parser.add_argument(
        "--size",
        "-S",
        action="append",
        help="Only include files of specific sizes (uses the same syntax as fd-find)",
    )
    parser.add_argument("--depth", "-d", default=0, type=int, help="Depth of folders")
    parser.add_argument("--include", "-s", nargs="+", action="extend", default=[], help=argparse.SUPPRESS)
    parser.add_argument("--exclude", "-E", "-e", nargs="+", action="extend", default=[], help=argparse.SUPPRESS)

    parser.add_argument("--tui", "-tui", action="store_true")
    parser.add_argument("--verbose", "-v", action="count", default=0)

    parser.add_argument("database")
    parser.add_argument("working_directory", nargs="*", default=os.sep)
    args = parser.parse_intermixed_args()
    args.db = db.connect(args)

    args.include += args.working_directory
    if args.include == ["."]:
        args.include = [str(Path().cwd().resolve())]

    if args.size:
        args.size = utils.parse_human_to_sql(utils.human_to_bytes, "size", args.size)

    log.info(utils.dict_filter_bool(args.__dict__))
    return args


def sort_by(args):
    if args.sort_by:
        return lambda x: x.get(args.sort_by)

    return lambda x: (x["size"] / (x.get("count") or 1), x["size"], x.get("count") or 1)


def get_subset(args, level=None, prefix=None) -> List[Dict]:
    d = {}
    excluded_files = set()

    for m in args.data:
        if prefix is not None and not m["path"].startswith(prefix):
            continue

        p = m["path"].split(os.sep)
        if level is not None and len(p) == level and not m["path"].endswith(os.sep):
            d[m["path"]] = m

        while len(p) >= 2:
            p.pop()
            if p == [""]:
                continue

            parent = os.sep.join(p) + os.sep
            if level is not None and len(p) != level:
                excluded_files.add(parent)

            if parent not in d:
                d[parent] = {"size": 0, "count": 0}
            d[parent]["size"] += m.get("size", 0)
            d[parent]["count"] += 1

    reverse = True
    if args.sort_by and " desc" in args.sort_by:
        args.sort_by = args.sort_by.replace(" desc", "")
        reverse = False

    return sorted(
        [{"path": k, **v} for k, v in d.items() if k not in excluded_files],
        key=sort_by(args),
        reverse=reverse,
    )


def load_subset(args):
    level = args.depth
    if args.depth == 0:
        while len(args.subset) < 2:
            level += 1
            args.subset = get_subset(args, level=level, prefix=args.cwd)
    else:
        args.subset = get_subset(args, level=level, prefix=args.cwd)

    if not args.subset:
        utils.no_media_found()

    args.cwd = os.sep.join(args.subset[0]["path"].split(os.sep)[: level - 1]) + os.sep
    return args.cwd, args.subset


def get_data(args) -> List[dict]:
    m_columns = db.columns(args, "media")
    args.filter_sql = []
    args.filter_bindings = {}

    if args.size:
        args.filter_sql.append(" and size IS NOT NULL " + args.size)
    db.construct_search_bindings(args, m_columns)

    media = list(
        args.db.query(
            f"""
        SELECT
            path
            , size
        FROM media m
        WHERE 1=1
            and size > 0
            {'and coalesce(time_deleted, 0) = 0' if 'time_deleted' in m_columns else ''}
            {'and coalesce(is_dir, 0) = 0' if 'is_dir' in m_columns else ''}
            {" ".join(args.filter_sql)}
        ORDER BY path
        """,
            args.filter_bindings,
        ),
    )

    if not media:
        utils.no_media_found()
    return media


def run_tui(args):
    from textual.app import App, ComposeResult
    from textual.containers import VerticalScroll
    from textual.reactive import var
    from textual.widgets import Footer, Label, OptionList

    class DiskUsage(App):
        CSS = """
        Screen {
            opacity: 0%;
        }

        Label {
            text-style: reverse;
        }

        """

        BINDINGS = [
            ("q", "quit", "Quit"),
            ("Left", "pop", "Go up"),
        ]

        def __init__(self, args):
            super().__init__()
            self.args = args
            self.dark = False
            var(self.args.subset, always_update=True)

        @staticmethod
        def _path(cwd, d):
            file_count = ""
            if d.get("count"):
                file_count = str(d["count"]) + " files"

            return "\t".join([d["path"].replace(cwd, ""), humanize.naturalsize(d["size"]), file_count])

        def compose(self) -> ComposeResult:
            yield Label("Disk Usage ~ Use the arrow keys to navigate")
            yield Label(f"--- {args.cwd} ---", id="cwd")
            yield VerticalScroll(OptionList(*[self._path(args.cwd, d) for d in self.args.subset]))
            yield Label(
                f"Total disk usage: {humanize.naturalsize(sum(d['size'] for d in self.args.subset))}  Items: {sum(d.get('count') or 1 for d in self.args.subset)}",
            )
            yield Footer()

        def on_mount(self) -> None:
            self.query_one(OptionList).focus()
            # self.call_later(self.load_directory, self.root)

        def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
            m = self.args.subset[event.option_index]
            if m.get("count"):
                self.args.cwd = m["path"]
                cwd, subset = load_subset(self.args)
                self.query_one(OptionList).clear_options()
                self.query_one(OptionList).add_options([self._path(cwd, d) for d in subset])
                self.query_one(OptionList).refresh()

            self.refresh()

        def action_pop(self) -> None:
            self.args.cwd.pop()
            load_subset(self.args)
            self.refresh()

    DiskUsage(args).run()


def disk_usage():
    args = parse_args()
    args.data = get_data(args)
    args.subset = []
    args.cwd = None

    load_subset(args)

    if args.tui:
        run_tui(args)
    else:
        tbl = deepcopy(args.subset)
        tbl = utils.col_naturalsize(tbl, "size")
        print(
            tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False, colalign=["left", "right"]),
        )


if __name__ == "__main__":
    disk_usage()



import itertools, os, sys
from inspect import isgenerator
from pathlib import Path
from types import ModuleType

import gallery_dl
from gallery_dl.exception import StopExtraction
from gallery_dl.extractor.message import Message
from gallery_dl.job import Job
from gallery_dl.util import build_duration_func

from xklb import consts, media, playlists, utils
from xklb.utils import log

gallery_dl = None


def load_module_level_gallery_dl(args) -> ModuleType:
    global gallery_dl

    if gallery_dl is None:
        import gallery_dl

        gallery_dl.config.load()

        download_archive = Path(args.download_archive or "~/.local/share/gallerydl.sqlite3").expanduser().resolve()
        if download_archive.exists():
            gallery_dl.config.set(("extractor",), "archive", str(download_archive))

        if hasattr(args, "prefix"):
            gallery_dl.config.set(("extractor",), "base-directory", args.prefix)
        gallery_dl.config.set(("extractor",), "parent-directory", False)
        gallery_dl.config.set(("extractor",), "directory", ["{category}", "{blog['name'][:30]}"])
        gallery_dl.config.set(
            ("extractor",),
            "filename",
            "{post['title'][:60]|summary[:30]|path[:50]}.{filename[:52]}.{extension}",
        )
        gallery_dl.config.set(("extractor",), "browser", "firefox")

        if consts.PYTEST_RUNNING:
            gallery_dl.config.set(("extractor",), "download", False)

    return gallery_dl


def is_supported(args, url) -> bool:
    if getattr(is_supported, "extractors", None) is None:
        gallery_dl = load_module_level_gallery_dl(args)
        is_supported.extractors = gallery_dl.extractor.extractors()

    return any(ie.pattern.match(url) and ie.category != "generic" for ie in is_supported.extractors)


def parse_gdl_job_status(job_status, path):
    errors = []

    if job_status & 1:
        errors.append("UnspecifiedError")
        log.error("[%s]: gallery_dl gave an UnspecifiedError.", path)

    if job_status & 64:  # no extractor
        errors.append("NoExtractorError")
        log.info("[%s]: NoExtractorError", path)

    if job_status & 16 or job_status & 32:
        if job_status & 16:
            errors.append("AuthorizationError")
        if job_status & 32:
            errors.append("FilterError")
        log.info("[%s]: Recoverable error(s) matched (will try again later)", path)

    if job_status & 4 or job_status & 8:  # HTTPError; not found / 404
        # TODO: distinguish between 429 and other errors
        # depending on extractor this can be fixed upstream
        # or by reading gallery_dl log stream
        error = "HTTPError"
        if job_status & 8:
            error = "HTTPNotFoundError"
        errors.append(error)
        log.debug("[%s]: Unrecoverable error %s. %s", path, error, utils.combine(errors))

    if job_status & 128:
        raise OSError
    if job_status & 2:
        raise ValueError("gallery_dl configuration error")

    return errors


def download(args, m):
    gallery_dl = load_module_level_gallery_dl(args)

    extensions = []
    if args.photos:
        extensions.extend(["jpg", "jpeg", "webp"])
    if args.drawings:
        extensions.extend(["png", "svg", "webp"])
    if args.gifs:
        extensions.extend(["gif", "mp4", "m4v", "webm"])

    if extensions:
        quoted_extensions = ",".join(f"'{ext}'" for ext in extensions)
        gallery_dl.config.set(("extractor",), "image-filter", f"extension in ({quoted_extensions})")

    webpath = m["path"]

    try:
        job = gallery_dl.job.DownloadJob(webpath)
    except gallery_dl.exception.NoExtractorError:
        log.info("[%s]: NoExtractorError", webpath)  # RecoverableError
        media.download_add(args, webpath, error="NoExtractorError")
        return

    job_status = job.run()
    errors = parse_gdl_job_status(job_status, webpath)

    info = getattr(job.pathfmt, "kwdict", None)
    if info:
        info["path"] = webpath

    local_path = getattr(job.pathfmt, "path", "") or None
    media.download_add(
        args,
        webpath,
        info,
        local_path=local_path,
        error=utils.combine(errors),
        unrecoverable_error="HTTPNotFoundError" in errors,
    )


class GeneratorJob(Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dispatched = False
        self.visited = set()
        self.status = 0

    def message_generator(self):
        extractor = self.extractor
        sleep = build_duration_func(extractor.config("sleep-extractor"))
        if sleep:
            extractor.sleep(sleep(), "extractor")

        try:
            for msg in extractor:
                self.dispatch(msg)
                if self.dispatched:
                    yield msg
                    self.dispatched = False
        except StopExtraction:
            pass

    def run(self):
        for msg in self.message_generator():
            ident, url, kwdict = msg
            if ident == Message.Url:
                yield (msg[1], msg[2])

            elif ident == Message.Queue:
                if url in self.visited:
                    continue
                self.visited.add(url)

                cls = kwdict.get("_extractor")
                if cls:
                    extr = cls.from_url(url)
                else:
                    extr = self.extractor.find(url)

                if extr:
                    job = self.__class__(extr, self)
                    for webpath, info in job.run():
                        yield (webpath, info)
            else:
                raise TypeError

    def handle_url(self, url, kwdict):
        self.dispatched = True

    def handle_queue(self, url, kwdict):
        self.dispatched = True


def get_playlist_metadata(args, playlist_path):
    gallery_dl = load_module_level_gallery_dl(args)

    added_media_count = 0
    job = GeneratorJob(playlist_path)
    gen = job.run()

    first_two = list(itertools.islice(gen, 2))
    is_playlist = len(first_two) > 1
    for webpath, info in itertools.chain(first_two, gen):
        errors = parse_gdl_job_status(job.status, playlist_path)
        extractor_key = job.extractor.category

        if not info:
            log.error("No info returned from image extractor %s", extractor_key)
            raise

        assert webpath
        info["extractor_key"] = extractor_key

        log.debug("webpath == playlist_path" if webpath == playlist_path else "webpath != playlist_path")

        playlist_id = None
        if is_playlist:
            playlist_id = playlists.add(args, playlist_path, info)
        else:
            log.warning("Importing playlist-less media %s", playlist_path)

        if media.exists(args, webpath):
            log.warning("Media already exists")

        info = {**info, "playlist_id": playlist_id, "webpath": webpath, **args.extra_media_data}
        media.playlist_media_add(
            args,
            playlist_path,
            info,
            error=utils.combine(errors),
            unrecoverable_error="HTTPNotFoundError" in errors,
        )

        added_media_count += 1
        if added_media_count > 1:
            sys.stdout.write("\033[K\r")
            print(f"[{playlist_path}] Added {added_media_count} media", end="\r", flush=True)

    if added_media_count == 0:
        from rich import inspect

        job = gallery_dl.job.DataJob(playlist_path, file=open(os.devnull, "w"), ensure_ascii=False)
        job.run()
        inspect(job.data)
        raise RuntimeError("No data found")

    return added_media_count



import argparse, json, tempfile
from copy import deepcopy
from pathlib import Path
from typing import List

from tabulate import tabulate

from xklb import consts, db, usage, utils
from xklb.utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="library redownload",
        usage=usage.redownload,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--download-archive", default="~/.local/share/yt_archive.txt")
    parser.add_argument("--limit", "-L", "-l", "-queue", "--queue", default=100)
    parser.add_argument("--verbose", "-v", action="count", default=0)

    parser.add_argument("database")
    parser.add_argument("deleted_at", nargs="?")
    parser.add_argument("deleted_to", nargs="?")
    args = parser.parse_args()
    args.db = db.connect(args)
    log.info(utils.dict_filter_bool(args.__dict__))
    return args


def list_deletions(args) -> List[dict]:
    query = """
        SELECT
            strftime('%Y-%m-%dT%H:%M:%S', time_deleted, 'unixepoch', 'localtime') as time_deleted
            , COUNT(*) as count
        FROM media
        WHERE time_deleted > 0
            AND time_downloaded > 0
        GROUP BY time_deleted
        HAVING count > 0
        ORDER BY time_deleted DESC
        LIMIT ?
    """
    media = list(args.db.query(query, [args.limit]))
    media = list(reversed(media))
    return media


def get_non_tube_media(args, paths) -> List[dict]:
    media = []
    paths = utils.conform(paths)
    if paths:
        for p in utils.chunks(paths, consts.SQLITE_PARAM_LIMIT):
            with args.db.conn:
                media.extend(
                    list(
                        args.db.query(
                            "select * from media where path in (" + ",".join(["?"] * len(p)) + ")",
                            (*p,),
                        ),
                    ),
                )
    return media


def get_deleted_media(args) -> List[dict]:
    if all([args.deleted_at, args.deleted_to]):
        # use timestamps between inclusive, converting from localtime to UTC
        query = """
            SELECT *
            FROM media
            WHERE time_deleted >= strftime('%s', ?, 'utc') AND time_deleted <= strftime('%s', ?, 'utc')
            AND time_downloaded > 0
        """
        media = list(args.db.query(query, (args.deleted_at, args.deleted_to)))
    else:
        # use exact timestamp, converting from localtime to UTC
        query = """
            SELECT *
            FROM media
            WHERE time_deleted = strftime('%s', ?, 'utc')
            AND time_downloaded > 0
        """
        media = list(args.db.query(query, (args.deleted_at,)))
    return media


def mark_media_undownloaded(args, deleted_media) -> None:
    m_columns = db.columns(args, "media")

    media = deepcopy(deleted_media)
    for d in media:
        d["time_deleted"] = 0
        d["time_modified"] = 0
        d["time_downloaded"] = 0
        d.pop("error", None)

        if "webpath" in m_columns and (d.get("webpath") or "").startswith("http"):
            with args.db.conn:
                args.db.conn.execute("DELETE from media WHERE path = ?", d["path"])

            d["path"] = d["webpath"]
            args.db["media"].upsert(d, pk="id", alter=True)  # type: ignore
        else:
            args.db["media"].upsert(d, pk="id", alter=True)  # type: ignore


def print_deletions(args, deletions) -> None:
    print("Deletions:")
    tbl = deepcopy(deletions)
    print(tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False))
    print(f"Showing most recent {args.limit} deletions. Use -l to change this limit")


def print_deleted(args, deleted_media) -> None:
    tbl = deepcopy(deleted_media)
    tbl = utils.list_dict_filter_bool(tbl, keep_0=False)
    tbl = utils.list_dict_filter_unique(tbl)
    tbl = tbl[: int(args.limit)]
    tbl = utils.col_resize(tbl, "path", 25)
    tbl = utils.col_duration(tbl, "duration")
    tbl = utils.col_naturalsize(tbl, "size")
    for t in consts.TIME_COLUMNS:
        utils.col_naturaldate(tbl, t)
    print(tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False))
    print(f"{len(deleted_media)} deleted media found (showing first {args.limit})")


def redownload() -> None:
    args = parse_args()

    if args.deleted_at:
        deleted_media = get_deleted_media(args)
    else:
        deletions = list_deletions(args)
        print_deletions(args, deletions)
        raise SystemExit(0)

    print_deleted(args, deleted_media)
    paths = [d["path"] for d in deleted_media]
    redownload_ids = [d["extractor_id"] for d in deleted_media if d.get("extractor_id")]
    print(len(redownload_ids), "tube ids found")
    if deleted_media and utils.confirm("Redownload media?"):  # type: ignore
        if len(redownload_ids) > 0:
            download_archive = Path(args.download_archive).expanduser().resolve()
            if download_archive.exists():
                utils.filter_file(str(download_archive), redownload_ids)

        mark_media_undownloaded(args, deleted_media)
        non_tube_media = get_non_tube_media(args, paths)

        print("Marked", len(deleted_media) - len(non_tube_media), "records as downloadable. Redownload via lb download")
        if len(non_tube_media) > 0:
            try:
                import pandas as pd

                out_path = tempfile.mktemp(".csv")
                pd.DataFrame(non_tube_media).to_csv(out_path, index=False)
            except ModuleNotFoundError:
                out_path = tempfile.mktemp(".json")
                with open(out_path, "w") as jf:
                    json.dump(non_tube_media, jf)
            print(
                len(non_tube_media),
                "records not recognized as tube media, which you will need to redownload manually. Exported to this temp file:",
                out_path,
            )


if __name__ == "__main__":
    redownload()



import argparse, shutil, tempfile
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Tuple

import humanize
from tabulate import tabulate

from xklb import consts, db, player, usage, utils
from xklb.utils import log


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="library mv-list",
        usage=usage.mv_list,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--limit", "-L", "-l", "-queue", "--queue", default="25")
    parser.add_argument("--lower", default=4, type=int, help="Number of files per folder lower limit")
    parser.add_argument("--upper", default=4000, type=int, help="Number of files per folder upper limit")
    parser.add_argument("--verbose", "-v", action="count", default=0)

    parser.add_argument("mount_point")
    parser.add_argument("database")
    args = parser.parse_args()
    args.db = db.connect(args)
    log.info(utils.dict_filter_bool(args.__dict__))
    return args


def group_by_folder(args, media) -> List[Dict]:
    d = {}
    for m in media:
        if m["path"].startswith("http"):
            continue

        p = m["path"].split("/")
        while len(p) >= 3:
            p.pop()
            parent = "/".join(p) + "/"

            if d.get(parent):
                d[parent]["size"] += m["size"]
                d[parent]["count"] += 1
            else:
                d[parent] = {
                    "size": m["size"],
                    "count": 1,
                }

    for path, pdict in list(d.items()):
        if any([pdict["count"] < args.lower, pdict["count"] > args.upper]):
            d.pop(path)

    return [{**v, "path": k} for k, v in d.items()]


def get_table(args) -> List[dict]:
    media = list(
        args.db.query(
            """
        select
            path
            , size
        from media
        where 1=1
            and time_deleted = 0
        order by path
        """,
        ),
    )

    folders = group_by_folder(args, media)
    return sorted(folders, key=lambda x: x["size"] / x["count"])


def iterate_and_show_options(args, tbl) -> Tuple[List[Dict], List[Dict]]:
    vew = tbl[-int(args.limit) :] if args.limit else tbl

    vew = utils.list_dict_filter_bool(vew, keep_0=False)
    vew = utils.col_resize(vew, "path", 60)
    vew = utils.col_naturalsize(vew, "size")
    print(tabulate(vew, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False))
    print(len(tbl) - len(vew), "other folders not shown")

    if args.limit:
        return tbl[-int(args.limit) :], tbl[: -int(args.limit)]
    else:
        return tbl, tbl


def move_list() -> None:
    args = parse_args()
    _total, _used, free = shutil.disk_usage(args.mount_point)

    print("Current free space:", humanize.naturalsize(free))

    data = get_table(args)

    tbl = deepcopy(data)
    cur, rest = iterate_and_show_options(args, tbl)

    data = {d["path"]: d for d in data}

    utils.set_readline_completion(list(data.keys()))

    print(
        """
██╗███╗░░██╗░██████╗████████╗██████╗░██╗░░░██╗░█████╗░████████╗██╗░█████╗░███╗░░██╗░██████╗
██║████╗░██║██╔════╝╚══██╔══╝██╔══██╗██║░░░██║██╔══██╗╚══██╔══╝██║██╔══██╗████╗░██║██╔════╝
██║██╔██╗██║╚█████╗░░░░██║░░░██████╔╝██║░░░██║██║░░╚═╝░░░██║░░░██║██║░░██║██╔██╗██║╚█████╗░
██║██║╚████║░╚═══██╗░░░██║░░░██╔══██╗██║░░░██║██║░░██╗░░░██║░░░██║██║░░██║██║╚████║░╚═══██╗
██║██║░╚███║██████╔╝░░░██║░░░██║░░██║╚██████╔╝╚█████╔╝░░░██║░░░██║╚█████╔╝██║░╚███║██████╔╝
╚═╝╚═╝░░╚══╝╚═════╝░░░░╚═╝░░░╚═╝░░╚═╝░╚═════╝░░╚════╝░░░░╚═╝░░░╚═╝░╚════╝░╚═╝░░╚══╝╚═════╝░

Type "done" when finished
Type "more" to see more files
Paste a folder (and press enter) to toggle selection
Type "*" to select all files in the most recently printed table
""",
    )

    selected_paths = set()
    while True:
        try:
            input_path = input("Paste a path: ").strip()
        except EOFError:
            break
        if input_path.lower() in ["done", "q"]:
            break

        if input_path.lower() == "more":
            cur, rest = iterate_and_show_options(args, rest)
            continue

        if input_path == "*":
            if cur:
                selected_paths.update([d["path"] for d in cur])
            else:
                selected_paths.update(data.keys())

            cur, rest = iterate_and_show_options(args, rest)
        else:
            try:
                data[input_path]
            except KeyError:
                continue

            if input_path in selected_paths:
                selected_paths.discard(input_path)
            else:
                selected_paths.add(input_path)

        # remove child paths so that the size of data is not counted twice
        temp_set = selected_paths.copy()
        for path1 in temp_set:
            for path2 in temp_set:
                if path1 != path2 and path1.startswith(path2):
                    selected_paths.discard(path1)

        selected_paths_size = sum([data[p]["size"] for p in selected_paths])
        print(
            len(selected_paths),
            "selected paths:",
            humanize.naturalsize(selected_paths_size),
            "; future free space:",
            humanize.naturalsize(selected_paths_size + free),
        )

    if selected_paths:
        temp_file = Path(tempfile.mktemp())
        with temp_file.open("w") as f:
            f.writelines("\n".join(selected_paths))

        print(
            f"""

    Folder list saved to {temp_file}. You may want to use the following command to move files to an EMPTY folder target:

        rsync -aE --xattrs --info=progress2 --no-inc-recursive --remove-source-files --files-from={temp_file} -r -vv --dry-run / jim:/free/real/estate/
        """,
        )

        if utils.confirm(f"Mark as deleted in {args.database}?"):  # type: ignore
            player.mark_media_deleted_like(args, list(selected_paths))


if __name__ == "__main__":
    move_list()



import argparse, sqlite3
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from dateutil import parser

from xklb import consts, db, fs_extract, utils
from xklb.consts import DBType
from xklb.utils import combine, log, safe_unpack


def exists(args, path) -> bool:
    m_columns = db.columns(args, "media")
    try:
        known = args.db.execute(
            f"select 1 from media where path=? or {'web' if 'webpath' in m_columns else ''}path=?",
            [str(path), str(path)],
        ).fetchone()
    except sqlite3.OperationalError as e:
        log.debug(e)
        return False
    if known is None:
        return False
    return True


def get(args, path):
    return args.db.pop_dict("select * from media where path = ?", [path])


def get_paths(args):
    tables = args.db.table_names()

    known_playlists = set()
    if "media" in tables:
        known_playlists.update(d["path"] for d in args.db.query("SELECT path from media"))

        m_columns = db.columns(args, "media")
        if "webpath" in m_columns:
            known_playlists.update(d["webpath"] for d in args.db.query("SELECT webpath from media"))

    if "playlists" in tables:
        known_playlists.update(d["path"] for d in args.db.query("SELECT path from playlists"))

    return known_playlists


def consolidate(v: dict) -> Optional[dict]:
    if v.get("title") in ("[Deleted video]", "[Private video]"):
        return None

    v = utils.flatten_dict(v, passthrough_keys=["automatic_captions", "http_headers", "subtitles"])

    upload_date = safe_unpack(
        v.pop("upload_date", None),
        v.pop("release_date", None),
        v.pop("date", None),
        v.pop("created_at", None),
        v.pop("published", None),
        v.pop("updated", None),
    )
    if upload_date:
        if isinstance(upload_date, datetime):
            upload_date = int(upload_date.replace(tzinfo=timezone.utc).timestamp())
        else:
            try:
                upload_date = int(parser.parse(upload_date).replace(tzinfo=timezone.utc).timestamp())
            except Exception:
                upload_date = None

    cv = {}
    cv["playlist_id"] = v.pop("playlist_id", None)
    cv["webpath"] = safe_unpack(
        v.pop("webpath", None),
        v.pop("webpage_url", None),
        v.pop("url", None),
        v.pop("original_url", None),
        v.pop("post_url", None),
        v.pop("image_permalink", None),
    )
    cv["path"] = safe_unpack(v.pop("path", None), v.pop("local_path", None), cv["webpath"])
    size_bytes = safe_unpack(v.pop("filesize_approx", None), v.pop("size", None))
    cv["size"] = 0 if not size_bytes else int(size_bytes)
    duration = v.pop("duration", None)
    cv["duration"] = 0 if not duration else int(duration)
    cv["time_uploaded"] = upload_date
    cv["time_created"] = consts.APPLICATION_START
    cv["time_modified"] = 0  # this should be 0 if the file has never been downloaded
    cv["time_deleted"] = 0
    cv["time_downloaded"] = 0
    language = v.pop("language", None)
    cv["tags"] = combine(
        "language:" + language if language else None,
        v.pop("description", None),
        v.pop("caption", None),
        v.pop("content", None),
        v.pop("categories", None),
        v.pop("genre", None),
        v.pop("tags", None),
        v.pop("labels", None),
    )

    cv["latitude"] = safe_unpack(
        v.pop("latitude", None),
        v.pop("lat", None),
    )
    cv["longitude"] = safe_unpack(
        v.pop("longitude", None),
        v.pop("long", None),
        v.pop("lng", None),
    )

    # extractor_key should only be in playlist table
    cv["extractor_id"] = v.pop("id", None)
    cv["title"] = safe_unpack(v.pop("title", None), v.get("playlist_title"))
    cv["width"] = v.pop("width", None)
    cv["height"] = v.pop("height", None)
    fps = v.pop("fps", None)
    cv["fps"] = 0 if not fps else int(fps)
    cv["live_status"] = v.pop("live_status", None)
    cv["age_limit"] = safe_unpack(
        v.pop("age_limit", None),
        18 if v.pop("is_mature", None) or v.pop("is_nsfw", None) else 0,
    )

    account = v.pop("account", None) or {}
    cv["uploader"] = safe_unpack(
        v.pop("channel_id", None),
        v.pop("uploader_url", None),
        v.pop("channel_url", None),
        v.pop("uploader", None),
        v.pop("channel", None),
        v.pop("uploader_id", None),
        account.pop("username", None),
        v.pop("account_id", None),
        account.pop("id", None),
        v.pop("name", None),
        v.pop("author", None),
        v.pop("post_author", None),
        v.pop("blog_name", None),
        v.pop("uuid", None),
        v.pop("playlist_uploader_id", None),
        v.pop("playlist_uploader", None),
    )

    cv["view_count"] = utils.safe_unpack(
        v.pop("view_count", None),
    )
    cv["num_comments"] = utils.safe_unpack(
        v.pop("replies", None),
    )
    cv["favorite_count"] = utils.safe_sum(
        v.pop("favorite_count", None),
        v.pop("likes", None),
        v.pop("note_count", None),
        v.pop("point_count", None),
    )
    cv["score"] = safe_unpack(v.pop("score", None))
    cv["upvote_ratio"] = v.pop("average_rating", None)
    if v.get("upvote_count") and v.get("upvote_count"):
        upvote_count = v.pop("upvote_count")
        downvote_count = v.pop("downvote_count")
        cv["upvote_ratio"] = upvote_count / (upvote_count + downvote_count)

    v = {
        k: v
        for k, v in v.items()
        if not (k.startswith(("_", "reblogged_")) or k in consts.MEDIA_KNOWN_KEYS or v is None)
    }
    if v != {}:
        log.info("Extra media data %s", v)
        # breakpoint()

    return utils.dict_filter_bool(cv)


def add(args, entry):
    if "path" not in entry:
        log.warning('Skipping insert: no "path" in entry %s', entry)
        return

    tags = entry.pop("tags", None) or ""

    media_id = args.db.pop("select id from media where path = ?", [entry["path"]])
    try:
        if media_id:
            entry["id"] = media_id

            args.db["media"].upsert(utils.dict_filter_bool(entry), pk="id", alter=True)
        else:
            args.db["media"].insert(utils.dict_filter_bool(entry), pk="id", alter=True)
            media_id = args.db.pop("select id from media where path = ?", [entry["path"]])
    except sqlite3.IntegrityError:
        log.error("media_id %s: %s", media_id, entry)
        raise
    if tags:
        args.db["captions"].insert({"media_id": media_id, "time": 0, "text": tags}, alter=True)


def playlist_media_add(
    args,
    webpath: str,
    info: Optional[dict] = None,
    error=None,
    unrecoverable_error=False,
) -> None:
    if not info:
        info = {"path": webpath}

    consolidated_entry = consolidate(info) or {}

    entry = {
        **consolidated_entry,
        "time_deleted": consts.APPLICATION_START if unrecoverable_error else 0,
        "webpath": webpath,
        "error": error,
    }
    add(args, entry)


def download_add(
    args,
    webpath: str,
    info: Optional[dict] = None,
    local_path=None,
    error=None,
    unrecoverable_error=False,
) -> None:
    if local_path and Path(local_path).exists():
        local_path = str(Path(local_path).resolve())
        fs_args = argparse.Namespace(
            profile=args.profile,
            scan_subtitles=args.profile == DBType.video,
            ocr=False,
            speech_recognition=False,
            delete_unplayable=False,
            check_corrupt=0.0,
            delete_corrupt=None,
        )
        fs_tags = utils.dict_filter_bool(fs_extract.extract_metadata(fs_args, local_path), keep_0=False) or {}
        fs_extract.clean_up_temp_dirs()
        with args.db.conn:
            args.db.conn.execute("DELETE from media WHERE path = ?", [webpath])
    else:
        fs_tags = {"time_modified": consts.now()}

    if not info:  # not downloaded or already downloaded
        info = {"path": webpath}

    consolidated_entry = consolidate(info) or {}

    entry = {
        **consolidated_entry,
        "time_deleted": consts.APPLICATION_START if unrecoverable_error else 0,
        **fs_tags,
        "webpath": webpath,
        "error": error,
    }
    add(args, entry)



import argparse, sys

import humanize
from tabulate import tabulate

from xklb import consts, db, player, tube_backend, usage, utils
from xklb.consts import SC
from xklb.utils import log


def parse_args():
    parser = argparse.ArgumentParser(
        prog="library block",
        usage=usage.block,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--print", "-p", default=False, const="p", nargs="?", help=argparse.SUPPRESS)
    parser.add_argument("--limit", "-L", "-l", "-queue", "--queue", help=argparse.SUPPRESS)
    parser.add_argument("--match-column", "-c", default="path", help="Column to block media if text matches")

    parser.add_argument("--force", "-f", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--offline", "--no-tube", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--ignore-errors", "--ignoreerrors", "-i", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--db", "-db", help=argparse.SUPPRESS)
    parser.add_argument("--verbose", "-v", action="count", default=0)

    parser.add_argument("database", help=argparse.SUPPRESS)
    parser.add_argument("playlists", nargs="+", help=argparse.SUPPRESS)
    args = parser.parse_intermixed_args()

    if args.db:
        args.database = args.db
    args.db = db.connect(args)

    args.playlists = utils.conform(args.playlists)

    args.action = SC.block
    log.info(utils.dict_filter_bool(args.__dict__))

    return args


def add_to_blocklist(args, p):
    p = p.pop()
    args.db["blocklist"].insert({"key": args.match_column, "value": p}, alter=True, replace=True, pk=["key", "value"])


def block(args=None) -> None:
    if args:
        sys.argv = ["lb", *args]

    args = parse_args()
    m_columns = db.columns(args, "media")
    if args.match_column not in m_columns:
        raise ValueError(
            "Match column does not exist in the media table. You may need to run tubeadd first or check your spelling",
        )

    columns = set(["path", "webpath", args.match_column, "size", "playlist_path", "time_deleted"])
    select_sql = ", ".join(s for s in columns if s in m_columns)

    if args.match_column == "playlist_path":
        playlist_paths = list(args.playlists)
        for p in playlist_paths:
            add_to_blocklist(args, [p])

        with args.db.conn:
            args.db.conn.execute(
                f"""UPDATE playlists
                SET time_deleted={consts.APPLICATION_START}
                WHERE path IN ("""
                + ",".join(["?"] * len(playlist_paths))
                + ")",
                (*playlist_paths,),
            )
        with args.db.conn:
            args.db.conn.execute(
                f"""UPDATE media
                SET time_deleted={consts.APPLICATION_START}
                WHERE path LIKE 'http%'
                AND playlist_id in (
                    SELECT id from playlists
                    WHERE path IN ("""
                + ",".join(["?"] * len(playlist_paths))
                + "))",
                (*playlist_paths,),
            )

        playlist_media = list(
            args.db.query(
                """SELECT path, size FROM media
                WHERE time_deleted = 0
                AND time_downloaded > 0
                AND playlist_id in (
                    SELECT id from playlists
                    WHERE path IN ("""
                + ",".join(["?"] * len(playlist_paths))
                + "))",
                (*playlist_paths,),
            ),
        )
        total_size = sum(d["size"] for d in playlist_media)
        paths_to_delete = [d["path"] for d in playlist_media]
        if paths_to_delete:
            print(paths_to_delete)
            if utils.confirm(
                f"Would you like to delete these {len(paths_to_delete)} local files ({humanize.naturalsize(total_size)})?",
            ):
                player.delete_media(args, paths_to_delete)
        return

    unmatched_playlists = []
    for p in args.playlists:
        p = [p]
        if consts.PYTEST_RUNNING or args.force:
            add_to_blocklist(args, p)
            continue

        matching_media = list(
            args.db.query(f"select {select_sql} from media where {args.match_column} LIKE ?", (p[0],)),
        )

        if not matching_media:
            matching_media = list(args.db.query(f"select {select_sql} from media where path = ?", (p[0])))
            if matching_media:
                log.debug("tube: found local %s", matching_media)

        if args.match_column in ("playlist_path", "path") and not args.offline and not matching_media:
            data = tube_backend.get_video_metadata(args, p[0])
            if data:
                if args.match_column not in data:
                    log.warning("[%s]: Match column %s not found in tube metadata", p[0], args.match_column)
                    msg = ""
                    for key in data:
                        preview = str(data[key]).replace("\n", " ")
                        if len(preview) > 80:
                            preview = preview[:77] + "..."
                        msg += f"  {key}: {preview}\n"
                    log.warning(msg)
                else:
                    p[1] = data[args.match_column]
                    if p[1]:
                        matching_media = list(
                            args.db.query(f"select {select_sql} from media where {args.match_column} = ?", (p[1],)),
                        )

        if not matching_media:
            unmatched_playlists.append(p)
            continue

        try:
            matching_media = list(reversed(utils.cluster_dicts(args, matching_media)))
        except ModuleNotFoundError:
            pass

        tbl = utils.list_dict_filter_bool(matching_media)
        tbl = [
            {
                "title_path": "\n".join(utils.concat(d.get("title"), d.get("webpath"), d["path"])),
                **d,
            }
            for d in tbl
        ]
        tbl = [{k: v for k, v in d.items() if k not in ("title", "path", "webpath")} for d in tbl]
        tbl = utils.col_resize(tbl, "title_path", 40)
        tbl = utils.col_naturalsize(tbl, "size")
        tbl = utils.col_naturaldate(tbl, "time_deleted")
        print(tabulate(tbl, tablefmt=consts.TABULATE_STYLE, headers="keys", showindex=False))
        if utils.confirm(f"{len(matching_media)} media matching {p}. Add to blocklist?"):
            add_to_blocklist(args, p)
        else:
            continue

        paths_to_delete = [
            d["path"] for d in matching_media if d["time_deleted"] == 0 and not d["path"].startswith("http")
        ]
        if paths_to_delete:
            total_size = sum(d["size"] for d in matching_media if d["time_deleted"] == 0)
            if utils.confirm(
                f"Would you like to delete these {len(paths_to_delete)} local files ({humanize.naturalsize(total_size)})?",
            ):
                player.delete_media(args, paths_to_delete)

    if unmatched_playlists:
        log.error("Could not find media matching these URLs/words (rerun with --force to add a blocking rule):")
        log.error("  " + " ".join(t[0] for t in unmatched_playlists))
